#
# Tcl package index file
#
if {[package vsatisfies [package provide Tcl] 9.0-]} {
    package ifneeded tdom 0.9.6 \
        "[list load [file join $dir .. .. libtcl9tdom096.so.0.0]];
         [list source [file join $dir tdom.tcl]]"
} else {
    package ifneeded tdom 0.9.6 \
        "[list load [file join $dir .. .. libtdom096.so.0.0]];
         [list source [file join $dir tdom.tcl]]"
}
